function [gnew,hnew,bnew,sigma_one,p_one,q_one,delta_vec,CM]=...
    bulge_chasing_routine(g,h,b,sigma,p,q,m,defl_size, v_vec,n,CM,pol)
% uses implicit QR with compression to perform bulge chasing on the
% submatrix m:defl_size;
% uses double shift;

%
% (last update Jan 25, 2012)

newF=zeros(3,3,n-1);
newV=zeros(2,2,n-1);

% ****** Let us apply Theorem 4.1 ******

 %n=defl_size;
sigma_one=zeros(n-1,1);
delta_vec=zeros(n-2,1);
Qnm1=zeros(2);
Q=zeros(3,3,n-2); % unitary matrices will be stored here, the last one is 2x2
hU=zeros(4,n); % uncompressed generators, note that order is 4
gU=zeros(n,4); 
bU=zeros(4,4,n-1);
p_one=zeros(n,1); % updated vectors for rank-one correction
q_one=zeros(n,1);



% *** Parts 1 and 2: compute the Q's and update generators ***
% (new uncompressed generators are computed along with the Q's)

% ****** 1.1 *****
    
%G=hh3(v_vec);
G=hh3_new(v_vec);

if m>1
       G=eye(3);
end

Q(:,:,1)=G';   
theta=q(1:2).'; % 2x2 row vector
auxA=[g(1,:)*h(:,1),g(1,:)*b(:,:,1)*h(:,2),g(1,:)*b(:,:,1)*b(:,:,2);...
    sigma(1),g(2,:)*h(:,2),g(2,:)*b(:,:,2); p(3)*q(1),sigma(2),g(3,:)];
auxB=G*auxA;
gamma=auxB(1,1:2);
w=auxB(1,3:4);
f=auxB(2:3,1:2);
v=auxB(2:3,3:4);

auxC=[f,v*h(:,3); p(4)*theta, sigma(3)];
auxD=auxC*G';
s=auxD(1:2,1);
eta=auxD(1:2,2:3);
rho=auxD(3,1);
beta=auxD(3,2:3);

auxp=G*p(1:3);
p_one(1)=auxp(1);
c=auxp(2:3);
auxq=(q(1:3).')*G';
q_one(1)=auxq(1);
theta=auxq(2:3);

auxE1=auxC-[c;p(4)]*q(1:3).';
auxE2=auxE1*G';

% compute updated generators (of order 4)
hQ=Q(1:2,1,1);
bQ=Q(1:2,2:3,1);
dQ=Q(3,1,1);
gQ=Q(3,2:3,1);

gU(1,:)=[gamma, w];
hU(:,1)=[hQ;h(:,3)*dQ];
bU(:,:,1)=[bQ,zeros(2);h(:,3)*gQ,b(:,:,3)];


% ****** 1.2 ******

for k=2:n-3
    auxvec=auxE2(:,1);
    %auxvec=[s-c*q_one(k-1);rho-p(k+2)*q_one(k-1)];
%    G=hh3(auxvec);
     G=hh3_new(auxvec);
%    G=hhbulge(auxE2);
    auxvec2=G*auxvec;
    
    if k==defl_size-1
        auxvec(3)=0;
        G=[planerot(auxvec(1:2)),zeros(2,1);zeros(1,2),1];
        auxvec2=G*auxvec;
    end
    
    if k>=defl_size
        G=eye(3);
        auxvec2=auxvec;
    end
   
   if k<m
       G=eye(3);
       auxvec2=auxvec;
   end
   
   if k==m
       G=hh3_new(v_vec);
       auxvec2=G*auxvec;
   end

    Q(:,:,k)=G';
    delta=auxvec2(1);
    
    auxA=[eta,v*b(:,:,k+1);beta,g(k+2,:)];
    auxB=G*auxA;
    gamma=auxB(1,1:2);
    w=auxB(1,3:4);
    f=auxB(2:3,1:2);
    v=auxB(2:3,3:4);
    
    auxC=[f,v*h(:,k+2); p(k+3)*theta,sigma(k+2)];
    auxD=auxC*G';
    s=auxD(1:2,1);
    eta=auxD(1:2,2:3);
    rho=auxD(3,1);
    beta=auxD(3,2:3);
    
    auxp=G*[c;p(k+2)];
    p_one(k)=auxp(1);
    c=auxp(2:3);
    
     auxE1=auxC-[c;p(k+3)]*[theta,q(k+2)];
auxE2=auxE1*G';
    
    auxq=[theta,q(k+2)]*G';
    q_one(k)=auxq(1);
    theta=auxq(2:3);
    
    sigma_one(k-1)=delta+p_one(k)*q_one(k-1);
    delta_vec(k-1)=delta;
    
    % compute updated generators (of order 4)
    hQ=Q(1:2,1,k);
    bQ=Q(1:2,2:3,k);
    dQ=Q(3,1,k);
    gQ=Q(3,2:3,k);

    gU(k,:)=[gamma, w];
    hU(:,k)=[hQ;h(:,k+2)*dQ];
    bU(:,:,k)=[bQ,zeros(2);h(:,k+2)*gQ,b(:,:,k+2)];
    
end


% ****** 1.3 ******

%auxvec=[s-c*q_one(n-3);rho-p(n)*q_one(n-3)];
auxvec=auxE2(:,1);
G=hh3_new(auxvec);

if (n-2)==defl_size-1
        auxvec(3)=0;
        G=[planerot(auxvec(1:2)),zeros(2,1);zeros(1,2),1];        
end

    if n-2>=defl_size
        G=eye(3);
    end
  
   if (n-2)==m
       %G=hh3(v_vec);
       G=hh3_new(v_vec);
   end

Q(:,:,n-2)=G';
auxvec2=G*auxvec;
delta=auxvec2(1);

auxA=[eta,v*b(:,:,n-1);beta,g(n,:)];
auxB=G*auxA;
gamma=auxB(1,1:2);
w=auxB(1,3:4);
f=auxB(2:3,1:2);
v=auxB(2:3,3:4);

auxmatf=[f,v*h(:,n)]*G';
s=auxmatf(:,1);
eta=auxmatf(:,2:3);

auxp=G*[c;p(n)];
p_one(n-2)=auxp(1);
c=auxp(2:3);
auxq=[theta,q(n)]*G';
q_one(n-2)=auxq(1);
oldtheta=theta;
theta=auxq(2:3);
    
sigma_one(n-3)=delta+p_one(n-2)*q_one(n-3);
delta_vec(n-3)=delta;

% compute updated generators (of order 4)
hQ=Q(1:2,1,n-2);
bQ=Q(1:2,2:3,n-2);
dQ=Q(3,1,n-2);
gQ=Q(3,2:3,n-2);

gU(n-2,:)=[gamma, w];
hU(:,n-2)=[hQ;h(:,n)*dQ];

bU(:,1:2,n-2)=[bQ;h(:,n)*gQ];

% ****** 1.4 ******

% auxvec3=s-c*q_one(n-2);

auxfinalone=[f,v*h(:,n)]-c*[oldtheta,q(n)];
auxfinaltwo=auxfinalone*G';
auxvec3=auxfinaltwo(:,1);

[G,auxvec4]=planerot(auxvec3);

if defl_size<n
%if defl_size<n-1  % modified here
    G=eye(2);
    auxvec4=auxvec3;
end
delta=auxvec4(1);
Qnm1=G';

auxmatg=G*eta;
gamma=auxmatg(1,:);
f=auxmatg(2,:); % this is 1x2 now!
auxvec5=f*G';
s=auxvec5(1);  % scalar
eta=auxvec5(2);  % scalar

auxp=G*c;
p_one(n-1)=auxp(1);
p_one(n)=auxp(2);

auxq=theta*G';
q_one(n-1)=auxq(1);
q_one(n)=auxq(2);
    
sigma_one(n-2)=delta+p_one(n-1)*q_one(n-2);
delta_vec(n-2)=delta;
sigma_one(n-1)=s;

% complete generator update
hQ=Qnm1(:,1);
bQ=Qnm1(:,2);

%gU(n-1,:)=gamma;
gU(n-1,:)=[gamma,0,0];
%gU(n,:)=eta;
gU(n,:)=[eta,0,0,0];
%hU(:,n-1)=hQ;
hU(:,n-1)=[hQ;0;0];
%hU(:,n)=1;
hU(:,n)=[1;0;0;0];
%bU(:,:,n-1)=bQ;
bU(1:2,1,n-1)=bQ;

% ***** 2 ******
% the generator update procedure is included above


% *** CHECK: verify that the QR step has been performed correctly
% for k=1:n-2
%    CM(k:k+2,:)=Q(:,:,k)'*CM(k:k+2,:);
%    CM(:,k:k+2)=CM(:,k:k+2)*Q(:,:,k);
% end
% CM(n-1:n,:)=Qnm1'*CM(n-1:n,:);
% CM(:,n-1:n)=CM(:,n-1:n)*Qnm1;
% CMgen=compute_matrix(gU,hU,bU,sigma_one,p_one,q_one);
% errU=norm(triu(CM(1:defl_size,1:defl_size))-triu(CMgen(1:defl_size,1:defl_size)));
% fprintf('error with uncompressed generators: %g \n',errU)


% ****** 3 ******
% (compression according to Theorem 2.1)
hnew=zeros(2,n);
gnew=zeros(n,2);
bnew=zeros(2,2,n-1);

Sold=hU(:,n);
hnew(1,n)=1;
gnew(n,1)=gU(n,:)*Sold;
hnew(:,n-1)=[1;0];
bnew(:,:,n-1)=[0,0;1,0];
Snew=[hU(:,n-1), bU(:,:,n-1)*Sold];
gnew(n-1,:)=gU(n-1,:)*Snew;
z=[sigma_one(n-1), gnew(n,1)];  
xold=p_one(n);  
for k=n-1:-1:2
   [V,Y]=planerot([p_one(k);xold]);  % here V = V_k'
   newV(:,:,k)=V';
   xnew=Y(1);
   A=V*[sigma_one(k-1), gnew(k,:); xold*q_one(k-1), z];
   xold=xnew;
   auxvec=A(2,:)';
  
   [G1,Y1]=planerot(auxvec(2:3));
   [G2,Y2]=planerot([auxvec(1);Y1(1)]);
   F=[G2,[0;0];0,0,1]*[1,0,0;[0;0],G1];  
   %F=flipud(F);   % here F=F_k-1 (not *)
   F=fliplr(F'); % here F=F_k-1*
   F=F';
   
   newF(:,:,k-1)=F;
   hnew(:,k-1)=F(1:2,1);
   bnew(:,:,k-1)=F(1:2,2:3);
   Sold=Snew;
   Snew=hU(:,k-1)*(hnew(:,k-1))'+bU(:,:,k-1)*Sold*bnew(:,:,k-1)';
   z=A(1,1)*hnew(:,k-1)'+A(1,2:3)*bnew(:,:,k-1)';
   gnew(k-1,:)=gU(k-1,:)*Snew;
end
% newV(:,:,1)=eye(2);
% newF(:,:,n-1)=eye(3);
% % correggiamo la prima F (deve annullare tutto il bulge, non solo una riga)
% auxAA=A*F'; % the second row should be [0 0 1]
% auxF2=planerot([auxAA(1,2)';auxAA(1,1)']);
% %auxF3=[auxF2' zeros(2,1); zeros(1,2),1];
% %newF(:,:,1)=auxF3*newF(:,:,1);
% 
% auxF2=fliplr(flipud(auxF2'));
% 
% %[auxF2,Y]=planerot([p_one(1);xold]);  % if used change ' in def of auxF3
% auxAAA=auxAA(:,1:2);
% auxAAAA=auxAAA*auxF2;
% %if auxAAAA(1,2)<0   % a change of sign is needed (det = -1)
% %    auxF2(2,:)=-auxF2(2,:);
% %end
% 
% auxF3=[auxF2' zeros(2,1); zeros(1,2),1];
% auxF3(1,:)=-auxF3(1,:);   % attenzione!!! (e' un gesto disperato, lo so...)
% newF(:,:,1)=auxF3*newF(:,:,1);

%auxA=[A(2,:)',A(1,:)'];
%   [QQ,RR]=qr(auxA);
%   F=fliplr(QQ);
%   F=F';
%   newF(:,:,1)=F;

% CHECK if the V*F representation is correct
% VVV=compV(newV,n);
% FFF=compF(newF,n);
% AAA=VVV*FFF-p_one*q_one.';
% AD=AbsDistance(eig(AAA),roots(pol));
% display(AD)
% if AD>1e-10
%     fprintf('wrong VF representation \n')
% end


%*** CHECK if compression works:
% CMcomp=compute_matrix(gnew,hnew,bnew,sigma_one,p_one,q_one);
% err1=norm(CMgen-CMcomp);
% fprintf('error with compressed generators: %g \n',err1)

dummy=1;